/*
 * Name:    d_trans.c 
 *
 * Purpose: MX motor driver to move 1 or more motors in the same direction.
 *
 *---------------------------------------------------------------------------
 *
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "mx_util.h"
#include "mx_drivr.h"
#include "mx_motor.h"
#include "d_trans.h"

/* Initialize the motor driver jump table. */

MX_RECORD_FUNCTION_LIST mxd_trans_motor_record_function_list = {
	mxd_trans_motor_initialize_type,
	mxd_trans_motor_create_record_structures,
	mxd_trans_motor_finish_record_initialization,
	mxd_trans_motor_delete_record,
	mxd_trans_motor_print_motor_structure,
	mxd_trans_motor_read_parms_from_hardware,
	mxd_trans_motor_write_parms_to_hardware,
	mxd_trans_motor_open,
	mxd_trans_motor_close
};

MX_MOTOR_FUNCTION_LIST mxd_trans_motor_motor_function_list = {
	mxd_trans_motor_motor_is_busy,
	mxd_trans_motor_move_absolute,
	mxd_trans_motor_get_position,
	mxd_trans_motor_set_position,
	mxd_trans_motor_soft_abort,
	mxd_trans_motor_immediate_abort,
	mxd_trans_motor_positive_limit_hit,
	mxd_trans_motor_negative_limit_hit,
	mxd_trans_motor_find_home_position
};

/* Translation motor data structures. */

MX_RECORD_FIELD_DEFAULTS mxd_trans_motor_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_ANALOG_MOTOR_STANDARD_FIELDS,
	MX_MOTOR_STANDARD_FIELDS,
	MXD_TRANS_MOTOR_STANDARD_FIELDS
};

long mxd_trans_motor_num_record_fields
		= sizeof( mxd_trans_motor_record_field_defaults )
			/ sizeof( mxd_trans_motor_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_trans_motor_rfield_def_ptr
			= &mxd_trans_motor_record_field_defaults[0];

/* === */

#define NUM_TRANS_MOTOR_FIELDS	2

MX_EXPORT mx_status_type
mxd_trans_motor_initialize_type( long type )
{
        const char fname[] = "mxs_trans_motor_initialize_type()";

        const char field_name[NUM_TRANS_MOTOR_FIELDS][MXU_FIELD_NAME_LENGTH+1]
            = {
                "motor_record_array",
		"position_array" };

        MX_DRIVER *driver;
        MX_RECORD_FIELD_DEFAULTS *record_field_defaults;
        MX_RECORD_FIELD_DEFAULTS **record_field_defaults_ptr;
        MX_RECORD_FIELD_DEFAULTS *field;
        int i;
        long num_record_fields;
	long referenced_field_index;
        long num_motors_varargs_cookie;
        mx_status_type status;

        driver = mx_get_driver_by_type( mx_type_list, type );

        if ( driver == (MX_DRIVER *) NULL ) {
                return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
                        "Record type %ld not found in mx_type_list.", type );
        }

        record_field_defaults_ptr = driver->record_field_defaults_ptr;

        if (record_field_defaults_ptr == (MX_RECORD_FIELD_DEFAULTS **) NULL) {
                return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
                "'record_field_defaults_ptr' for record type '%s' is NULL.",
                        driver->name );
        }

        record_field_defaults = *record_field_defaults_ptr;

        if ( record_field_defaults == (MX_RECORD_FIELD_DEFAULTS *) NULL ) {
                return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
                "'record_field_defaults_ptr' for record type '%s' is NULL.",
                        driver->name );
        }

        if ( driver->num_record_fields == (long *) NULL ) {
                return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
                "'num_record_fields' pointer for record type '%s' is NULL.",
                        driver->name );
        }

	num_record_fields = *(driver->num_record_fields);

        status = mx_find_record_field_defaults_index(
                        record_field_defaults, num_record_fields,
                        "num_motors", &referenced_field_index );

        if ( status.code != MXE_SUCCESS )
                return status;

        status = mx_construct_varargs_cookie(
                        referenced_field_index, 0, &num_motors_varargs_cookie);

        if ( status.code != MXE_SUCCESS )
                return status;

        MX_DEBUG( 2,("%s: num_motors varargs cookie = %ld",
                        fname, num_motors_varargs_cookie));

	for ( i = 0; i < NUM_TRANS_MOTOR_FIELDS; i++ ) {
		status = mx_find_record_field_defaults(
			record_field_defaults, num_record_fields,
			field_name[i], &field );

		if ( status.code != MXE_SUCCESS )
			return status;

		field->dimension[0] = num_motors_varargs_cookie;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_trans_motor_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_trans_motor_create_record_structures()";

	MX_MOTOR *motor;
	MX_TRANSLATION_MOTOR *trans_motor;

	/* Allocate memory for the necessary structures. */

	motor = (MX_MOTOR *) malloc( sizeof(MX_MOTOR) );

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_MOTOR structure." );
	}

	trans_motor = (MX_TRANSLATION_MOTOR *)
				malloc( sizeof(MX_TRANSLATION_MOTOR) );

	if ( trans_motor == (MX_TRANSLATION_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_TRANSLATION_MOTOR structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = motor;
	record->record_type_struct = trans_motor;
	record->class_specific_function_list
				= &mxd_trans_motor_motor_function_list;

	motor->record = record;

	/* A translation motor is treated as an analog motor. */

	motor->subclass = MXC_MTR_ANALOG;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_trans_motor_finish_record_initialization( MX_RECORD *record )
{
	mx_status_type status;

	status = mx_motor_finish_record_initialization( record );

	return status;
}

MX_EXPORT mx_status_type
mxd_trans_motor_delete_record( MX_RECORD *record )
{
	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}
	if ( record->record_type_struct != NULL ) {
		free( record->record_type_struct );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_trans_motor_print_motor_structure( FILE *file, MX_RECORD *record )
{
	const char fname[] = "mxd_trans_motor_print_motor_structure()";

	MX_MOTOR *motor;
	MX_TRANSLATION_MOTOR *translation_motor;
	MX_RECORD **motor_record_array;
	long i, num_motors;
	double position, deadband;
	mx_status_type status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer for record '%s' is NULL.", record->name );
	}

	translation_motor = (MX_TRANSLATION_MOTOR *)
					(record->record_type_struct);

	if ( translation_motor == (MX_TRANSLATION_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_TRANSLATION_MOTOR pointer for record '%s' is NULL.",
			record->name );
	}

	num_motors = translation_motor->num_motors;
	motor_record_array = translation_motor->motor_record_array;


	fprintf(file, "MOTOR parameters for motor '%s':\n", record->name);
	fprintf(file, "  Motor type           = TRANSLATION_MOTOR.\n\n");

	fprintf(file, "  name                 = %s\n", record->name);

	status = mx_motor_get_position( record, &position );

	if ( status.code != MXE_SUCCESS ) {
		return mx_error( MXE_FUNCTION_FAILED, fname,
			"Unable to read position of motor '%s'",
			record->name );
	}

	fprintf(file, "  position             = %g %s  (%g).\n",
		motor->position, motor->units, motor->raw_position.analog );
	fprintf(file, "  scale                = %g %s per unscaled "
							"translation unit.\n",
		motor->scale, motor->units);
	fprintf(file, "  offset               = %g %s.\n",
		motor->offset, motor->units);
	fprintf(file, "  backlash             = %g %s  (%g).\n",
		motor->backlash_correction, motor->units,
		motor->raw_backlash_correction.analog);
	fprintf(file, "  negative limit       = %g %s  (%g).\n",
		motor->negative_limit, motor->units,
		motor->raw_negative_limit.analog);
	fprintf(file, "  positive limit       = %g %s  (%g).\n",
		motor->positive_limit, motor->units,
		motor->raw_positive_limit.analog);

	deadband = motor->scale * motor->deadband.analog;

	fprintf(file, "  deadband             = %g %s  (%g).\n\n",
		deadband, motor->units,
		motor->deadband.analog );

	fprintf(file, "  num motors           = %ld\n", num_motors);
	fprintf(file, "  motor record array   = (");

	if ( motor_record_array == NULL ) {
		fprintf(file, "NULL");
	} else {
		for ( i = 0; i < num_motors; i++ ) {
			fprintf(file, "%s", motor_record_array[i]->name);

			if ( i != num_motors-1 ) {
				fprintf(file, ", ");
			}
		}
	}
	fprintf(file, ")\n");

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_trans_motor_read_parms_from_hardware( MX_RECORD *record )
{
	/* All saving of parameters is handled by the MX_MOTOR records
	 * this motor is derived from, so we need not do anything here.
	 */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_trans_motor_write_parms_to_hardware( MX_RECORD *record )
{
	/* All setting of parameters is handled by the MX_MOTOR records
	 * this motor is derived from, so we need not do anything here.
	 */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_trans_motor_open( MX_RECORD *record )
{
	/* Nothing to do. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_trans_motor_close( MX_RECORD *record )
{
	/* Nothing to do. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_trans_motor_motor_is_busy( MX_MOTOR *motor )
{
	const char fname[] = "mxd_trans_motor_motor_is_busy()";

	MX_TRANSLATION_MOTOR *translation_motor;
	MX_RECORD **motor_record_array;
	MX_RECORD *child_motor_record;
	long i, num_motors;
	int busy;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed NULL.");
	}

	translation_motor = (MX_TRANSLATION_MOTOR *)
					(motor->record->record_type_struct);

	if ( translation_motor == (MX_TRANSLATION_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_TRANSLATION_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	num_motors = translation_motor->num_motors;
	motor_record_array = translation_motor->motor_record_array;

	for ( i = 0; i < num_motors; i++ ) {
	
		child_motor_record = motor_record_array[i];

		if ( child_motor_record == (MX_RECORD *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"Motor record pointer for child motor '%s' for translation motor '%s' is NULL.",
			motor_record_array[i]->name,
			motor->record->name );
		}

		status = mx_motor_is_busy( child_motor_record, &busy );

		if ( status.code != MXE_SUCCESS )
			return status;

		/* If the child motor is busy, we are done. */

		if ( busy == TRUE ) {
			motor->busy = TRUE;

			return MX_SUCCESSFUL_RESULT;
		}
	}

	motor->busy = FALSE;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_trans_motor_move_absolute( MX_MOTOR *motor )
{
	const char fname[] = "mxd_trans_motor_move_absolute()";

	MX_TRANSLATION_MOTOR *translation_motor;
	MX_RECORD **motor_record_array;
	MX_RECORD *child_motor_record;
	double *position_array;
	long i, num_motors;
	double new_translation_position;
	double old_translation_position;
	double translation_position_difference;
	double position_sum;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed NULL.");
	}

	translation_motor = (MX_TRANSLATION_MOTOR *)
					(motor->record->record_type_struct);

	if ( translation_motor == (MX_TRANSLATION_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_TRANSLATION_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	num_motors = translation_motor->num_motors;
	motor_record_array = translation_motor->motor_record_array;
	position_array = translation_motor->position_array;

	new_translation_position = motor->raw_destination.analog;

	/* Check the limits on this request. */

	if ( new_translation_position > motor->raw_positive_limit.analog
	  || new_translation_position < motor->raw_negative_limit.analog )
	{
		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
	"Requested position %g is outside the allowed range of %g to %g",
		new_translation_position, motor->raw_negative_limit.analog,
			motor->raw_positive_limit.analog );
	}

	/* Get the old positions. */

	position_sum = 0.0;

	for ( i = 0; i < num_motors; i++ ) {
		child_motor_record = motor_record_array[i];

		status = mx_motor_get_position(
			child_motor_record, &position_array[i] );

		if ( status.code != MXE_SUCCESS )
			return status;

		position_sum += position_array[i];
	}

	old_translation_position = position_sum / (double) num_motors;

	translation_position_difference = new_translation_position
		- old_translation_position;

	/* Now compute the new positions. */

	for ( i = 0; i < num_motors; i++ ) {
		position_array[i] += translation_position_difference;
	}

	/* Perform the move. */

	status = mx_motor_array_move_absolute(
			num_motors, motor_record_array, position_array,
			MXF_MTR_NOWAIT );

	return status;
}

MX_EXPORT mx_status_type
mxd_trans_motor_get_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_trans_motor_get_position()";

	MX_RECORD **motor_record_array;
	MX_RECORD *child_motor_record;
	double *position_array;
	long i, num_motors;
	MX_TRANSLATION_MOTOR *translation_motor;
	double translation_position;
	double position_sum;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed NULL.");
	}

	translation_motor = (MX_TRANSLATION_MOTOR *)
					(motor->record->record_type_struct);

	if ( translation_motor == (MX_TRANSLATION_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_TRANSLATION_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	num_motors = translation_motor->num_motors;
	motor_record_array = translation_motor->motor_record_array;
	position_array = translation_motor->position_array;

	/* Get the current motor positions. */

	position_sum = 0.0;

	for ( i = 0; i < num_motors; i++ ) {
		child_motor_record = motor_record_array[i];

		status = mx_motor_get_position(
			child_motor_record, &position_array[i] );

		if ( status.code != MXE_SUCCESS )
			return status;

		position_sum += position_array[i];
	}

	translation_position = position_sum / (double) num_motors;

	motor->raw_position.analog = translation_position;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_trans_motor_set_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_trans_motor_set_position()";

	return mx_error( MXE_UNSUPPORTED, fname,
	"'set position' is not valid for a translation motor." );
}

MX_EXPORT mx_status_type
mxd_trans_motor_soft_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_trans_motor_soft_abort()";

	MX_RECORD **motor_record_array;
	MX_RECORD *child_motor_record;
	long i, num_motors;
	MX_TRANSLATION_MOTOR *translation_motor;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed NULL.");
	}

	translation_motor = (MX_TRANSLATION_MOTOR *)
					(motor->record->record_type_struct);

	if ( translation_motor == (MX_TRANSLATION_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_TRANSLATION_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	num_motors = translation_motor->num_motors;
	motor_record_array = translation_motor->motor_record_array;

	for ( i = 0; i < num_motors; i++ ) {
		child_motor_record = motor_record_array[i];

		status = mx_motor_soft_abort( child_motor_record );
	}


	return status;
}

MX_EXPORT mx_status_type
mxd_trans_motor_immediate_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_trans_motor_immediate_abort()";

	MX_RECORD **motor_record_array;
	MX_RECORD *child_motor_record;
	long i, num_motors;
	MX_TRANSLATION_MOTOR *translation_motor;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed NULL.");
	}

	translation_motor = (MX_TRANSLATION_MOTOR *)
					(motor->record->record_type_struct);

	if ( translation_motor == (MX_TRANSLATION_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_TRANSLATION_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	num_motors = translation_motor->num_motors;
	motor_record_array = translation_motor->motor_record_array;

	for ( i = 0; i < num_motors; i++ ) {
		child_motor_record = motor_record_array[i];

		status = mx_motor_immediate_abort( child_motor_record );
	}

	return status;
}

MX_EXPORT mx_status_type
mxd_trans_motor_positive_limit_hit( MX_MOTOR *motor )
{
	const char fname[] = "mxd_trans_motor_positive_limit_hit()";

	MX_RECORD **motor_record_array;
	MX_RECORD *child_motor_record;
	MX_TRANSLATION_MOTOR *translation_motor;
	long i, num_motors;
	int limit_hit;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed NULL.");
	}

	translation_motor = (MX_TRANSLATION_MOTOR *)
					(motor->record->record_type_struct);

	if ( translation_motor == (MX_TRANSLATION_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_TRANSLATION_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	num_motors = translation_motor->num_motors;
	motor_record_array = translation_motor->motor_record_array;

	limit_hit = FALSE;

	for ( i = 0; i < num_motors; i++ ) {
		child_motor_record = motor_record_array[i];

		status = mx_motor_positive_limit_hit(
					child_motor_record, &limit_hit );

		if ( status.code != MXE_SUCCESS )
			return status;

		if ( limit_hit == TRUE ) {
			(void) mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
				"Positive limit hit for motor '%s'",
				child_motor_record->name );

			break;             /* exit the for() loop */
		}
	}

	motor->positive_limit_hit = limit_hit;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_trans_motor_negative_limit_hit( MX_MOTOR *motor )
{
	const char fname[] = "mxd_trans_motor_negative_limit_hit()";

	MX_RECORD **motor_record_array;
	MX_RECORD *child_motor_record;
	MX_TRANSLATION_MOTOR *translation_motor;
	long i, num_motors;
	int limit_hit;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed NULL.");
	}

	translation_motor = (MX_TRANSLATION_MOTOR *)
					(motor->record->record_type_struct);

	if ( translation_motor == (MX_TRANSLATION_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_TRANSLATION_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	num_motors = translation_motor->num_motors;
	motor_record_array = translation_motor->motor_record_array;

	limit_hit = FALSE;

	for ( i = 0; i < num_motors; i++ ) {
		child_motor_record = motor_record_array[i];

		status = mx_motor_negative_limit_hit(
					child_motor_record, &limit_hit );

		if ( status.code != MXE_SUCCESS )
			return status;

		if ( limit_hit == TRUE ) {
			(void) mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
				"Negative limit hit for motor '%s'",
				child_motor_record->name );

			break;             /* exit the for() loop */
		}
	}

	motor->negative_limit_hit = limit_hit;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_trans_motor_find_home_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_trans_motor_find_home_position()";

	return mx_error( MXE_UNSUPPORTED, fname,
"Finding home for a translation motor is _not_ allowed.  Motor name = '%s'",
		motor->record->name );
}

