/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/* mx_var.c --- Support for generic variable superclass support. */

#include <stdio.h>
#include <string.h>

#include "mx_rec.h"
#include "mx_alloc.h"
#include "mx_drivr.h"
#include "mx_var.h"

MX_EXPORT mx_status_type
mx_variable_initialize_type( long record_type )
{
	const char fname[] = "mx_variable_initialize_type()";

	MX_DRIVER *driver;
	MX_RECORD_FIELD_DEFAULTS *record_field_defaults;
	MX_RECORD_FIELD_DEFAULTS **record_field_defaults_ptr;
	long num_record_fields;
	mx_status_type status;

	driver = mx_get_driver_by_type( mx_type_list, record_type );

	if ( driver == (MX_DRIVER *) NULL ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Record type %ld not found in mx_type_list.",
			record_type );
	}

	record_field_defaults_ptr = driver->record_field_defaults_ptr;

	if (record_field_defaults_ptr == (MX_RECORD_FIELD_DEFAULTS **) NULL) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"'record_field_defaults_ptr' for record type '%s' is NULL.",
			driver->name );
	}

	record_field_defaults = *record_field_defaults_ptr;

	if ( record_field_defaults == (MX_RECORD_FIELD_DEFAULTS *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"'record_field_defaults_ptr' for record type '%s' is NULL.",
			driver->name );
	}

	if ( driver->num_record_fields == (long *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"'num_record_fields' pointer for record type '%s' is NULL.",
			driver->name );
	}

	/**** Fix up the record fields common to all variable types. ****/

	num_record_fields = *(driver->num_record_fields);

	status = mx_variable_fixup_varargs_record_field_defaults(
			record_field_defaults, num_record_fields );

	if ( status.code != MXE_SUCCESS )
		return status;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_variable_fixup_varargs_record_field_defaults(
		MX_RECORD_FIELD_DEFAULTS *record_field_defaults_array,
		long num_record_fields )
{
	const char fname[] =
		"mx_variable_fixup_varargs_record_field_defaults()";

	MX_RECORD_FIELD_DEFAULTS *dimension_field, *value_field;
	long num_dimensions_field_index;
	long num_dimensions_varargs_cookie;
	long dimension_field_index;
	long dimension_element_cookie;
	long i;
	mx_status_type status;

	MX_DEBUG( 8,("%s invoked.", fname));

	/****
	 **** Construct and place the 'num_dimensions' varargs cookie.
	 ****/

	status = mx_find_record_field_defaults_index(
			record_field_defaults_array, num_record_fields,
			"num_dimensions", &num_dimensions_field_index );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mx_construct_varargs_cookie(
		num_dimensions_field_index, 0, &num_dimensions_varargs_cookie);

	if ( status.code != MXE_SUCCESS )
		return status;

	MX_DEBUG( 8,("%s: num_dimensions varargs cookie = %ld",
			fname, num_dimensions_varargs_cookie));

	/*---*/

	status = mx_find_record_field_defaults_index(
			record_field_defaults_array, num_record_fields,
			"dimension", &dimension_field_index );

	if ( status.code != MXE_SUCCESS )
		return status;

	dimension_field = &record_field_defaults_array[dimension_field_index];

	dimension_field->dimension[0] = num_dimensions_varargs_cookie;

	/*---*/

	status = mx_find_record_field_defaults(
			record_field_defaults_array, num_record_fields,
			"value", &value_field );

	if ( status.code != MXE_SUCCESS )
		return status;

	value_field->num_dimensions = num_dimensions_varargs_cookie;

	/****
	 **** Construct the varargs cookies for each of the entries in
	 **** the 'value' array.
	 ****/

	for ( i = 0; i < MXU_FIELD_MAX_DIMENSIONS; i++ ) {

		status = mx_construct_varargs_cookie(
			dimension_field_index, i, &dimension_element_cookie );

		value_field->dimension[i] = dimension_element_cookie;

		MX_DEBUG( 8,("%s: dimension[%ld] cookie = %ld",
			fname, i, dimension_element_cookie));
	}

	return MX_SUCCESSFUL_RESULT;
}

/* A note concerning the mx_send_variable() and mx_receive_variable()
 * functions below.  For some variable classes, the MX_VARIABLE_FUNCTION_LIST
 * pointer is NULL.  This means that no data transfer need take place.
 * This is also true if the 'send_variable' or 'receive_variable' function
 * pointers are NULL.  An example of this are "inline" variables which have
 * the master copy of the data in the database record itself and do not
 * need to transfer any data anywhere.
 */

MX_EXPORT mx_status_type
mx_send_variable( MX_RECORD *record )
{
	const char fname[] = "mx_send_variable()";

	MX_VARIABLE *variable;
	MX_VARIABLE_FUNCTION_LIST *variable_flist;
	mx_status_type ( *fptr ) ( MX_VARIABLE * );
	mx_status_type status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_RECORD pointer passed was NULL." );
	}

	variable_flist = (MX_VARIABLE_FUNCTION_LIST *)
				record->superclass_specific_function_list;

	if ( variable_flist == (MX_VARIABLE_FUNCTION_LIST *) NULL ) {
		/* This is _not_ an error. */

		return MX_SUCCESSFUL_RESULT;
	}

	fptr = variable_flist->send_value;

	if ( fptr == NULL ) {
		/* This is also _not_ an error. */

		return MX_SUCCESSFUL_RESULT;
	}

	variable = (MX_VARIABLE *) record->record_superclass_struct;

	if ( variable == (MX_VARIABLE *) NULL ) {
		/* This _is_ an error if we get here. */

		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_VARIABLE pointer for record '%s' is NULL.",
			record->name );
	}

	status = ( *fptr ) ( variable );

	return status;
}

MX_EXPORT mx_status_type
mx_receive_variable( MX_RECORD *record )
{
	const char fname[] = "mx_receive_variable()";

	MX_VARIABLE *variable;
	MX_VARIABLE_FUNCTION_LIST *variable_flist;
	mx_status_type ( *fptr ) ( MX_VARIABLE * );
	mx_status_type status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_RECORD pointer passed was NULL." );
	}

	variable_flist = (MX_VARIABLE_FUNCTION_LIST *)
				record->superclass_specific_function_list;

	if ( variable_flist == (MX_VARIABLE_FUNCTION_LIST *) NULL ) {
		/* This is _not_ an error. */

		return MX_SUCCESSFUL_RESULT;
	}

	fptr = variable_flist->receive_value;

	if ( fptr == NULL ) {
		/* This is also _not_ an error. */

		return MX_SUCCESSFUL_RESULT;
	}

	variable = (MX_VARIABLE *) record->record_superclass_struct;

	if ( variable == (MX_VARIABLE *) NULL ) {
		/* This _is_ an error if we get here. */

		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_VARIABLE pointer for record '%s' is NULL.",
			record->name );
	}

	status = ( *fptr ) ( variable );

	return status;
}

MX_EXPORT mx_status_type
mx_get_variable_parameters( MX_RECORD *record,
		long *num_dimensions,
		long **dimension_array,
		long *field_type )
{
	MX_RECORD_FIELD *value_field;
	mx_status_type status;

	status = mx_find_record_field( record, "value", &value_field );

	if ( status.code != MXE_SUCCESS )
		return status;

	*num_dimensions = value_field->num_dimensions;
	*dimension_array = value_field->dimension;
	*field_type = value_field->datatype;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_get_variable_pointer( MX_RECORD *record, void **variable_pointer )
{
	MX_RECORD_FIELD *value_field;
	mx_status_type status;

	void *data_ptr;

	status = mx_find_record_field( record, "value", &value_field );

	if ( status.code != MXE_SUCCESS )
		return status;

	data_ptr = value_field->data_pointer;

	if ( value_field->flags & MXFF_VARARGS ) {
		*variable_pointer
			= mx_read_void_pointer_from_memory_location(data_ptr );
	} else {
		*variable_pointer = data_ptr;
	}

	return MX_SUCCESSFUL_RESULT;
}

/*** Note that the following functions only work for 1-dimensional arrays. ***/

MX_EXPORT mx_status_type
mx_get_variable_with_type( MX_RECORD *record_list,
		char *record_name,
		long requested_field_type,
		long *num_elements,
		void **pointer_to_value )
{
	const char fname[] = "mx_get_variable_with_type()";

	MX_RECORD *record;
	long num_dimensions, *dimension_array, actual_field_type;
	mx_status_type status;

	record = mx_get_record( record_list, record_name );

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_UNDEFINED_RECORD, fname,
		"The record '%s' does not exist.", record_name );
	}

	status = mx_get_variable_parameters( record, &num_dimensions,
			&dimension_array, &actual_field_type );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( actual_field_type != requested_field_type ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
"The 'value' field for '%s' is not of type %s.  Instead, it is of type %s.",
			record_name,
			mx_get_field_type_string( requested_field_type ),
			mx_get_field_type_string( actual_field_type ) );
	}

	if ( num_dimensions != 1 ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
"%s only supports 1-dimensional arrays.  '%s.value' has %ld dimensions.",
			fname, record_name, num_dimensions );
	}

	*num_elements = dimension_array[0];

	status = mx_get_variable_pointer( record, pointer_to_value );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Fetch the value of the variable from a remote server, if any. */

	status = mx_receive_variable( record );

	return status;
}

MX_EXPORT mx_status_type
mx_get_variable_from_record( MX_RECORD *record,
		long requested_field_type,
		long *num_elements,
		void **pointer_to_value )
{
	const char fname[] = "mx_get_variable_from_record()";

	long num_dimensions, *dimension_array, actual_field_type;
	mx_status_type status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"MX_RECORD pointer passed was NULL." );
	}

	status = mx_get_variable_parameters( record, &num_dimensions,
			&dimension_array, &actual_field_type );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( actual_field_type != requested_field_type ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
"The 'value' field for '%s' is not of type %s.  Instead, it is of type %s.",
			record->name,
			mx_get_field_type_string( requested_field_type ),
			mx_get_field_type_string( actual_field_type ) );
	}

	if ( num_dimensions != 1 ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
"%s only supports 1-dimensional arrays.  '%s.value' has %ld dimensions.",
			fname, record->name, num_dimensions );
	}

	*num_elements = dimension_array[0];

	status = mx_get_variable_pointer( record, pointer_to_value );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Fetch the value of the variable from a remote server, if any. */

	status = mx_receive_variable( record );

	return status;
}

MX_EXPORT mx_status_type
mx_get_long_variable( MX_RECORD *record_list,
		char *record_name,
		long *long_value )
{
	const char fname[] = "mx_get_long_variable()";

	long num_elements;
	void *pointer_to_value;
	mx_status_type status;

	if ( long_value == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"long_value pointer passed was NULL." );
	}

	status = mx_get_variable_with_type( record_list, record_name,
			MXFT_LONG, &num_elements, &pointer_to_value );

	if ( status.code != MXE_SUCCESS )
		return status;

	*long_value = *( (long *) pointer_to_value );

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_get_unsigned_long_variable( MX_RECORD *record_list,
		char *record_name,
		unsigned long *unsigned_long_value )
{
	const char fname[] = "mx_get_unsigned_long_variable()";

	long num_elements;
	void *pointer_to_value;
	mx_status_type status;

	if ( unsigned_long_value == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"unsigned_long_value pointer passed was NULL." );
	}

	status = mx_get_variable_with_type( record_list, record_name,
			MXFT_ULONG, &num_elements, &pointer_to_value );

	if ( status.code != MXE_SUCCESS )
		return status;

	*unsigned_long_value = *( (unsigned long *) pointer_to_value );

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_get_double_variable( MX_RECORD *record_list,
		char *record_name,
		double *double_value )
{
	const char fname[] = "mx_get_double_variable()";

	long num_elements;
	void *pointer_to_value;
	mx_status_type status;

	if ( double_value == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"double_value pointer passed was NULL." );
	}

	status = mx_get_variable_with_type( record_list, record_name,
			MXFT_DOUBLE, &num_elements, &pointer_to_value );

	if ( status.code != MXE_SUCCESS )
		return status;

	*double_value = *( (double *) pointer_to_value );

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_get_string_variable( MX_RECORD *record_list,
		char *record_name,
		char *string_value,
		size_t max_string_length )
{
	const char fname[] = "mx_get_string_variable()";

	long num_elements;
	void *pointer_to_value;
	char *pointer_to_string;
	mx_status_type status;

	if ( string_value == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"string_value pointer passed was NULL." );
	}

	status = mx_get_variable_with_type( record_list, record_name,
			MXFT_STRING, &num_elements, &pointer_to_value );

	if ( status.code != MXE_SUCCESS )
		return status;

	pointer_to_string = (char *) pointer_to_value;

	strcpy( string_value, "" );

	strncat( string_value, pointer_to_string, max_string_length );

	return MX_SUCCESSFUL_RESULT;
}

